-- during alpha, every release will increment MAJOR to 0.x,
-- since there is no presumption of backwards compatibility
local MAJOR,MINOR = "LibOOP-0.1",1
local LibOOP,oldminor = LibStub:NewLibrary(MAJOR, MINOR)

if not LibOOP then return end -- already loaded

-- localize global functions
local assert,error,getmetatable,setmetatable,type = assert,error,getmetatable,setmetatable,type

-- define utility tables
local weakKeys = { __mode = 'k' }
--local weakVals = { __mode = 'v' }
--local weakTable = { __mode = 'kv' }

-- define storage for class infrastructure (indexed by class table)
local ooBase = setmetatable({}, weakKeys)
local ooSuper = setmetatable({}, weakKeys)
--local ooClassMeta = setmetatable({}, weakKeys) --needed for finalize()
local ooProto = setmetatable({}, weakKeys)
--local ooProtoMeta = setmetatable({}, weakKeys) --needed for finalize()
local ooObjectMeta = setmetatable({}, weakKeys)

-- define placeholders for the base class (so methods can access them in closure)
local baseClass = {}
local baseProto = {}
ooSuper[baseClass] = true
ooProto[baseClass] = baseProto


--[[ UTILITY FUNCTIONS ]]--

local function noop() end
local function final() error("cannot modify finalized class") end
local function abstractNew() error("cannot instantiate abstract class") end


--[[ CLASS METHODS ]]--

local function class_new(class, ...)
	local meta = ooObjectMeta[class]
	assert(meta)
	local obj = setmetatable({}, meta)
	obj:__init(...)
	return obj
end -- class_new()

local function class_extend(class) -- requires class_new
	assert(ooSuper[class])
	local newClassMeta = {
		__index = class,
		__call = class_new,
		__metatable = false
	}
	local newClass = setmetatable({}, newClassMeta)
	local newProtoMeta = {
		__index = ooProto[class],
		__metatable = false
	}
	local newProto = setmetatable({}, newProtoMeta)
	local newObjectMeta = {
		__index = newProto,
		__metatable = newClass
	}
	newClass.prototype = newProto
	ooSuper[newClass] = class
	--ooClassMeta[newClass] = newClassMeta
	ooProto[newClass] = newProto
	--ooProtoMeta[newClass] = newProtoMeta
	ooObjectMeta[newClass] = newObjectMeta
	return newClass	
end -- class_extend()

-- don't call this on baseClass
local function class_extends(class, super, direct)
	local classSuper = ooSuper[class]
	if (not classSuper) then
		return false
	elseif (not super or classSuper == super) then
		return true
	elseif (direct or not ooSuper[super]) then
		return false
	end
	classSuper = ooSuper[classSuper]
	while classSuper do
		if (classSuper == super) then
			return true
		end
		classSuper = ooSuper[classSuper]
	end
	return false
end -- class_extends()

local function class_super(class, func, ...)
	local parent = ooSuper[class]
	assert(parent and func and type(parent[func])=="function")
	return parent[func](class, ...)
end -- class_super()

local function class_superclass(class)
	local parent = ooSuper[class]
	if (not parent or parent == baseClass) then
		return nil
	end
	return parent
end -- class_superclass()


--[[ OBJECT METHODS ]]--

local function obj_class(obj)
	local class = getmetatable(obj)
	if (type(ooSuper[class]) ~= "table") then
		return nil
	end
	return class
end -- obj_class()

local function obj_instanceof(obj, class, direct)
	local objClass = obj_class(obj)
	if (not ooSuper[objClass]) then
		return false
	elseif (not class or objClass == class) then
		return true
	elseif (direct or not ooSuper[class]) then
		return false
	end
	return class_extends(objClass, class)
end -- obj_instanceof()

local function obj_super(obj, func, ...)
	local class = obj_class(obj)
	local parent = ooSuper[class]
	assert(parent)
	local proto = ooProto[parent]
	assert(func and (type(proto[func])=="function"))
	return proto[func](obj, ...)
end -- obj_super()


--[[ BASE CLASS ]]--

baseClass.extend = class_extend
baseClass.extends = class_extends
baseClass.new = class_new
baseClass.super = class_super
baseClass.superclass = class_superclass

baseProto.__init = noop
baseProto.class = obj_class
baseProto.instanceof = obj_instanceof
baseProto.super = obj_super


--[[ LIBRARY API ]]--

--- Define a new class.
-- When extending an existing class (by passing a reference to it), this is the
-- same as "otherClass:extend()".  The new class will inherit any properties or
-- methods defined on superclasses, including the methods "extend", "new",
-- "extends" and "super" from the base class.  The new class will also provide
-- a "prototype" table on which properties and methods may be defined that will
-- be inherited by all objects of the class or its subclasses.
-- @param [super] Optional; another class to extend.
-- @return A table representing the newly defined class.
function LibOOP:Class(super)
	return class_extend(super or baseClass)
end -- Class()

--- Test if a reference is a class, or extends a specified class.
-- This is the same as "class:extends(super,direct)".  Only the first argument
-- is required; if the rest are omitted, this returns true if "class" is a
-- LibOOP class definition.  If "super" is given, this returns true if "class"
-- extends "super", even indirectly (i.e. "class" extends "other" which extends
-- "super").  If "direct" is true, then "class" must extend "super" directly,
-- with no intermediate superclass.
-- @param class The reference to test the inheritence of.
-- @param [super] Optional; a class to test inheritence from.
-- @param [direct] Optional; if true, tests only direct inheritence.
-- @return True if the class satisfies the specified inheritence profile, false otherwise.
function LibOOP:Extends(class, super, direct)
	return class_extends(class, super, direct)
end -- Extends()

--- Get the class of which the given object is an instance.
-- This is the same as "obj:class()".
-- @param obj The object to get the class of.
-- @return A reference to the class of the object.
function LibOOP:GetClass(obj)
	return obj_class(obj)
end -- GetClass()

--- Get the superclass of the given class.
-- This is the same as "class:superclass()".
-- @param class The class to get the superclass of.
-- @return A reference to the class' superclass.
function LibOOP:GetSuperClass(class)
	return class_superclass(class)
end -- GetSuperClass()

--- Test if a reference is an object, or is an instance of a specified class.
-- This is the same as "obj:instanceof(class,direct)".  Only the first argument
-- is required; if the rest are omitted, this returns true if "obj" is a LibOOP
-- object.  If "class" is given, this returns true if "obj" is an instance of
-- "class", even indirectly (i.e. "obj" is an instance of "other", which
-- extends "class").  If "direct" is true, then "obj" must be an instance of
-- "class" itself, and not a subclass of it.
-- @param obj The reference to test the instantiation of.
-- @param [class] Optional; a class to test instantiation from.
-- @param [direct] Optional; if true, tests only direct instantiation.
-- @return True if the object satisfies the specified instantiation profile, false otherwise.
function LibOOP:InstanceOf(obj, class, direct)
	return obj_instanceof(obj, class, direct)
end -- InstanceOf()

--- Create a new object of the given class.
-- This is the same as "class:new(...)" or "class(...)".  If the class
-- prototype defines an "__init" method, it will be called on the new object
-- using all extra arguments to this function, but its return value will be
-- ignored.  The new object will inherit any properties or methods defined on
-- the class' or any superclass' prototype, including the methods "class",
-- "instanceof" and "super" from the base prototype.
-- @param class The class to instantiate.
-- @return A table representing the newly created object.
function LibOOP:New(class, ...)
	return class_new(class, ...)
end -- New()

--- Call the inherited version of a function on the given class or object.
-- This is the same as "ref:super(func,...)".  If the reference is a class,
-- this will call "func" as known by the class' superclass; if the reference
-- is an object, it will call "func" as known by the object's class'
-- superclass' prototype.  In either case the remaining arguments are passed
-- to the function and the result returned.  This is useful when a class or
-- prototype has overridden a method inherited from its superclass, and the
-- redefinition of the method needs to call the inherited version.
-- @param ref A class or object reference.
-- @param func The name of the inherited function to be called, as a string.
-- @param ... Any number of arguments to pass to the function.
-- @return All return values from the function.
function LibOOP:Super(ref, func, ...)
	if (ooSuper[ref]) then
		return class_super(ref, func, ...)
	end
	return obj_super(ref, func, ...)
end -- Super()
